/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.TaskStepListener;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;

import java.util.*;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Verwaltet eine Warteschlange und eine Liste von Tasks. Ein Auftrag aus der Warteschlage wird dem naechsten freien Task zugewiesen. Wenn kein Task frei ist,
 * wartet der MultiTaskManager.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 7744 $ / $Date: 2010-06-01 19:35:01 +0200 (Di, 01 Jun 2010) $ / ($Author: rs $)
 */
public class MultiTaskManager extends Task implements RingOfDeathListener {

	protected RingBuffer<ResultData> inputDataQueue;

	protected RingBuffer<MultiTask> freeTaskQueue;

	protected ArrayList<MultiTask> taskList = new ArrayList<MultiTask>();

	/** So viele Tasks sind beim naechsten Durchlauf zu entfernen */
	private AtomicInteger adjustTaskNumFactor = new AtomicInteger(0);

	private Semaphore terminationSemaphor = null;

	public MultiTaskManager(ArchiveManager archiveMgr, RingOfDeathListener rodl, RingBuffer<ResultData> inputData, String taskName) {
		super(archiveMgr, rodl);
		logger = Debug.getLogger();
		setName(getName() + "[" + taskName + "]");
		freeTaskQueue = new RingBuffer<MultiTask>(16, RingBuffer.UNBOUNDED_SIZE);	// waechst/schrumpft blockweise um 16
		inputDataQueue = inputData;
		setPriority(9);
	}

	public void start() {
		for(MultiTask task : taskList) {
			task.start();
		}
		super.start();
	}

	/**
	 * Fuegt ein MultiTask hinzu und startet ihn, falls der MultitaskManager selbst schon gestartet ist.
	 *
	 * @param task
	 *
	 * @throws ConfigurationException
	 */
	public synchronized void addTask(MultiTask task) {
		if(isTaskTerminated()) return;

		synchronized(taskList) {
			taskList.add(task);
			task.setTaskIndex(taskList.size() - 1);
		}
		// nur starten wenn der MultitaskManager selbst schon gestartet ist:
		if(!(getState() == Thread.State.NEW || getState() == Thread.State.TERMINATED)) task.start();
	}

	/** Entfernt einen Subtask aus der Liste der verfuegbaren Tasks. Es verbleibt immer mindestens ein Task. */
	public synchronized void removeTask() {
		if(getTaskNum() > 1) adjustTaskNumFactor.incrementAndGet();		// ...und bei der naechsten Zuweisung ueberspringen
	}

	public int getTaskNum() {
		synchronized(taskList) {
			return taskList.size() - adjustTaskNumFactor.get();		// die zu entfernenden abziehen...
		}
	}

	/**
	 * Diese Methode wird von einem Task aufgerufen, wenn dieser Task mit der Bearbeitung eines Auftrages fertig ist und einen neuen Auftrag entgegennehmen kann.
	 *
	 * @param task
	 */
	public void pushFreeTask(MultiTask task) {
		if(!isTaskTerminated()) freeTaskQueue.push(task);
	}

	public void run() {
		try {
			while(!isTaskTerminated()) {
				try {
					ResultData rd = inputDataQueue.pop();			// synchronized auf inputDataQueue, blockiert falls leer

					MultiTask nextWorkingTask;
					while(true) {
						nextWorkingTask = freeTaskQueue.pop();		// synchronized auf freeTaskQueue, blockiert falls leer

						if(adjustTaskNumFactor.get() > 0) {			// Wenn Tasks entfernt werden sollen, werden sie hier
							adjustTaskNumFactor.decrementAndGet();	// uebersprungen und beendet.
							nextWorkingTask.terminateTask();
							synchronized(taskList) {
								taskList.remove(nextWorkingTask);	// aus der statischen Liste entfernen
							}
							continue;
						}

						if(!nextWorkingTask.isTaskTerminated())	// Keinem abgebrochenen Task eine Aufgabe zuweisen!!
						{
							break;
						}
					}
					nextWorkingTask.doWork(rd);						// nextWorkingTask aufwecken

					synchronized(taskStepListeners) {
						for(TaskStepListener tsl : taskStepListeners) {
							tsl.taskStepDone(this);
						}
					}
				}
				catch(InterruptedException e) {
					break;
				}
				catch(Exception e) {
					logger.error("Fehler", e);
				}
			}
			isTerminated = true;
			waitForSubtaskTermination();

			synchronized(taskList) {
				taskList.clear();			// taskList dient auch als mutex, drum nicht auf null setzen
				freeTaskQueue = null;
			}

		}
		finally {
			if(ringOfDeathListener != null) ringOfDeathListener.ringOfDeath(this);
			ringOfDeathListener = null;
		}
	}

	private void waitForSubtaskTermination() {
		synchronized(taskList) {
			terminationSemaphor = new Semaphore(taskList.size());
			try {
				terminationSemaphor.acquire(taskList.size());
				for(MultiTask task : taskList) {
					if(task.getState() == Thread.State.NEW) {
						terminationSemaphor.release();		// wenn nicht gestartet, Totenglckchen direkt aufrufen
					}
					else {
						task.terminateTask();
					}
				}
				// acquire(n) blockiert bis die gewuenschte Zahl n vorhanden ist, d.h. bis fuer jeden Task in 
				// ringOfDeath() ein Semaphor (mit release()) zurueckgegeben wurde.
				terminationSemaphor.acquire(taskList.size());
			}
			catch(InterruptedException e) {
			}	// ignorieren
		}
	}

	/* (non-Javadoc)
		 * @see de.bsvrz.ars.ars.mgmt.tasks.RingOfDeathListener#ringOfDeath(de.bsvrz.ars.ars.mgmt.tasks.Task)
		 */
	public synchronized void ringOfDeath(Task doomedTask) {
		logger.finer("Task beendet: " + doomedTask.getName());

		// Nur wenn der TaskManager beendet wurde, braucht es dan Semaphor. Nicht wenn der Subtask
		// sich von selber beendet hat:
		if(terminationSemaphor != null) terminationSemaphor.release();
	}

	public String toString() {
		StringBuilder sb = new StringBuilder(2048);
		synchronized(taskList) {
			for(MultiTask mt : taskList) {
				sb.append(mt != null ? (mt.getName() + ":  " + mt.getState()) : "null");
				sb.append("\n");
			}
		}
		return sb.toString();
	}
}
