/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.util.*;


/**
 * Klasse zum Durchlaufen des Persistenzverzeichnisses.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 6419 $ / $Date: 2009-03-10 23:15:57 +0100 (Di, 10 Mrz 2009) $ / ($Author: rs $)
 */
public class ContainerFileDirIterator {

	private static int MAX_DIR_DEPTH = 32;

	private ArrayList<File[]> tree = new ArrayList<File[]>(MAX_DIR_DEPTH);

	private ArrayList<Integer> pos = new ArrayList<Integer>(MAX_DIR_DEPTH);

	private int level = 0;

	/**
	 * Erzeugt einen neuen ContainerFileDirIterator.
	 *
	 * @param rootDir Wurzel des Persistenzverzeichnisses.
	 */
	public ContainerFileDirIterator(String rootDir) {
		addDirs(0, new File(rootDir).listFiles());
	}

	/**
	 * Liefert das naechste Verzeichnis, in dem eine Datenidentifikation gespeichert ist (d.h. "oa", "on", "na" oder "nn"). Es werden nur Verzeichnisse betrachtet,
	 * die den Namenskonventionen des {@link PersistenceManager} entsprechen.
	 *
	 * @return File-Objekt des Verzeichnisses oder <code>null</code>, wenn es keines mehr gibt.
	 */
	public File getNextDIdDirectory() {
		if(level < 0) return null;

		// am Ende eines Verzeichnisses: eins rauf, eins nach rechts
		if((getPos(level) >= getDirs(level).length)) {
			nextPos(--level);
			return getNextDIdDirectory();
		}

		File file = getDirs(level)[getPos(level)];

		// Treffer? eins nach rechts, Verzeichnis ausgeben
		if(PersistenceManager.isValidDataKindFolderName(file.getName())) {
			nextPos(level);
			return file;
		}

		// Ungueltige Verzeichnisnamen ueberspringen
		if(!PersistenceManager.isValidIDFolderName(file.getName())) {
			nextPos(level);
			return getNextDIdDirectory();
		}

		// Mittendrin:
		File[] nextDirs = Util.listDirectories(getDirs(level)[getPos(level)]);
		addDirs(++level, nextDirs);
		return getNextDIdDirectory();
	}

	/**
	 * Liefert das naechste Verzeichnis einer Simulationsvariante (d.h. sv<SimVar>). Es werden nur Verzeichnisse betrachtet, die den Namenskonventionen des {@link
	 * PersistenceManager} entsprechen. Wurde der Iterator durch {@link #ContainerFileDirIterator(String)} erzeugt, so werden die Verzeichnisse der
	 * Simulationsvariante <code>0</code> zurckgegeben.
	 *
	 * @return File-Objekt des Verzeichnisses oder <code>null</code>, wenn es keines mehr gibt.
	 * @see #ContainerFileDirIterator(String,short)
	 */
	public File getNextSimVarDirectory() {
		if(level < 0) return null;

		// am Ende eines Verzeichnisses: eins rauf, eins nach rechts
		if((getPos(level) >= getDirs(level).length)) {
			nextPos(--level);
			return getNextSimVarDirectory();
		}

		File file = getDirs(level)[getPos(level)];

		// Treffer? Verzeichnis ausgeben
		if(PersistenceManager.isValidSvFolderName(file.getName())) {
			nextPos(level);
			return file;
		}

		// Ungueltige Verzeichnisnamen ueberspringen
		if(!PersistenceManager.isValidIDFolderName(file.getName())) {
			nextPos(level);
			return getNextSimVarDirectory();
		}

		// Mittendrin:
		File[] nextDirs = Util.listDirectories(getDirs(level)[getPos(level)]);
		addDirs(++level, nextDirs);
		return getNextSimVarDirectory();
	}

	private int getPos(int lvl) {
		return pos.get(lvl);
	}

	private File[] getDirs(int lvl) {
		return tree.get(lvl);
	}

	private void addDirs(int lvl, File[] dirs) {
		if(lvl >= tree.size()) {
			tree.add(dirs);
			pos.add(0);
		}
		else {
			tree.set(lvl, dirs);
			pos.set(lvl, 0);
		}
	}

	private void nextPos(int lvl) {
		if(lvl >= 0) pos.set(lvl, pos.get(lvl) + 1);
	}
}


