/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTreeIterator;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.util.*;

/**
 * Klasse zum Iterieren ueber Containergrenzen einer Datenidentifikation hinweg. Es wird stets ueber aufsteigende Datenindexe iteriert.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 8179 $ / $Date: 2010-09-20 13:07:40 +0200 (Mo, 20 Sep 2010) $ / ($Author: rs $)
 */
public class MultiContainerDataIterator {

	private static final boolean DONT_INCLUDE_DELETED = false;

	private Debug logger;

	/** Datensatzart-Index. */
	private static int OA = 0, NN = 3;

	/** Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt. */
	private Task ownerTask;

	/** Archive-Manager fuer DataIdentTree-Zugriff. */
	private ArchiveManager arcMgr;

	/** Persistenz-Manager fuer Container-Datei-Zugriff. */
	private PersistenceManager persMgr;

	/** Persistent-Verzeichnis der aktuellen Datenidentifikation. */
	private File didDir;

	/** Liste mit Container-Walks fuer die einzelnen Datensatzarten. */
	private ContainerWalk[] contWalks = new ContainerWalk[4];

	/** Der Container-Walk, aus dem der letzte Datensatz geliefert worden ist. */
	private ContainerWalk lastContWalk = null;

	/** Geloeschte Container werden beim Iterieren ignoriert */
	public static final int DEL_IGNORE = 0;

	/** Beim Itererieren geloeschter Container werden nur 2 Datensaetze (Min/Max) gezaehlt */
	public static final int DEL_MINMAX = 1;

	/** Beim Iterieren geloeschter Container werden alle Datensaetze gezaehlt */
	public static final int DEL_FULLCOUNT = 2;

	/**
	 * Gibt an, wie ber geloeschte Container iteriert wird,<br> 0 : Container wird ignoriert <br> 1 : Container wird als 2 Datensaetze (Min/Max) gezaehlt<br> 2 :
	 * Container wird voll gezaehlt
	 */
	private int deletedTreatment = DEL_IGNORE;

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Konstruktoren
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 *
	 * @param arcMgr           Der Archiv-Manager.
	 * @param objID            Die Objekt-ID.
	 * @param atgID            Die Attributgruppen-ID.
	 * @param aspID            Die Aspekt-ID.
	 * @param sv               Die Simulationsvariante.
	 * @param oa               Kennzeichen, ob Container mit Online-Aktuell erhaltenen Datensaetzen genommen werden.
	 * @param on               Kennzeichen, ob Container mit Online-Nachgeliefert erhaltenen Datensaetzen genommen werden.
	 * @param na               Kennzeichen, ob Container mit Nachgefordert-Aktuell erhaltenen Datensaetzen genommen werden.
	 * @param nn               Kennzeichen, ob Container mit Nachgefordert-Nachgeliefert erhaltenen Datensaetzen genommen werden.
	 * @param deletedTreatment gibt an, wie geloeschte Container behandelt werden sollen.
	 * @param ownerTask        Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr,
			long objID,
			long atgID,
			long aspID,
			int sv,
			boolean oa,
			boolean on,
			boolean na,
			boolean nn,
			int deletedTreatment,
			Task ownerTask) throws PersistenceException {
		this.deletedTreatment = deletedTreatment;
		this.ownerTask = ownerTask;
		this.arcMgr = arcMgr;
		this.persMgr = arcMgr.getPersMgr();
		didDir = new File(persMgr.getPathName(objID, atgID, aspID, sv));

		boolean[] adks = new boolean[]{oa, on, na, nn};
		for(int archiveDataKindIndex = OA; archiveDataKindIndex <= NN; archiveDataKindIndex++) {
			if(adks[archiveDataKindIndex]) {
				ArchiveDataKind archiveDataKind = Util.getDataKindFromIndex(archiveDataKindIndex);
				final String pathname = persMgr.getPathName(objID, atgID, aspID, sv, archiveDataKind);
				//CacheManager.getInstance().createCachedFiles(pathname);
				File adkPath = new File(pathname);
				File[] contFiles = adkPath.listFiles(ContainerFile.CONT_FILENAME_FILTER);
				Util.sortFilesAlpha(contFiles); // Nach Namen (= Datenindexen, falls nicht NN) sortieren
				ContInfo[] contInfos = getContInfos(contFiles, archiveDataKind);
				ContainerWalk contWalk = new ContainerWalk(archiveDataKind);
				if(archiveDataKind.isRequested()) {
					contWalk.fillSorted(contInfos); // Nach Datenindex sortiert befuellen
				}
				else {
					contWalk.fillSimple(contInfos); // Nach Dateinamen = ContainerIndex = DatenIndex befuellen
				}
				contWalks[archiveDataKindIndex] = contWalk;
				getNextData(contWalk); // ContWalk haelt jetzt den ersten zu liefernden Datensatz bereit.
			}
		}
	}

	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 *
	 * @param arcMgr    Der Archiv-Manager.
	 * @param objID     Die Objekt-ID.
	 * @param atgID     Die Attributgruppen-ID.
	 * @param aspID     Die Aspekt-ID.
	 * @param sv        Die Simulationsvariante.
	 * @param oa        Kennzeichen, ob Container mit Online-Aktuell erhaltenen Datensaetzen genommen werden.
	 * @param on        Kennzeichen, ob Container mit Online-Nachgeliefert erhaltenen Datensaetzen genommen werden.
	 * @param na        Kennzeichen, ob Container mit Nachgefordert-Aktuell erhaltenen Datensaetzen genommen werden.
	 * @param nn        Kennzeichen, ob Container mit Nachgefordert-Nachgeliefert erhaltenen Datensaetzen genommen werden.
	 * @param ownerTask Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr, long objID, long atgID, long aspID, int sv, boolean oa, boolean on, boolean na, boolean nn, Task ownerTask)
			throws PersistenceException {
		this(arcMgr, objID, atgID, aspID, sv, oa, on, na, nn, DEL_IGNORE, ownerTask);
	}


	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 *
	 * @param arcMgr           Der Archiv-Manager.
	 * @param objID            Die Objekt-ID.
	 * @param atgID            Die Attributgruppen-ID.
	 * @param aspID            Die Aspekt-ID.
	 * @param sv               Die Simulationsvariante.
	 * @param adkc             Liefert Kennzeichen, von welchen Datensatzarten die Container genommen werden
	 * @param deletedTreatment Gibt an, ob geloeschte Container ignoriert werden [0], als 2 Datensaetze (Min/Max) zaehlen [1], oder voll gezaehlt werden [2]
	 * @param ownerTask        Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr, long objID, long atgID, long aspID, int sv, ArchiveDataKindCombination adkc, int deletedTreatment, Task ownerTask)
			throws PersistenceException {
		this(
				arcMgr,
				objID,
				atgID,
				aspID,
				sv,
				adkc.isOnline(),
				adkc.isOnlineDelayed(),
				adkc.isRequested(),
				adkc.isRequestedDelayed(),
				deletedTreatment,
				ownerTask
		);
	}

	/** @see #MultiContainerDataIterator(ArchiveManager,long,long,long,int,ArchiveDataKindCombination,int,Task) mit deletedTreatment==0 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr, long objID, long atgID, long aspID, int sv, ArchiveDataKindCombination adkc, Task ownerTask) throws PersistenceException {
		this(arcMgr, objID, atgID, aspID, sv, adkc.isOnline(), adkc.isOnlineDelayed(), adkc.isRequested(), adkc.isRequestedDelayed(), DEL_IGNORE, ownerTask);
	}

	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 * Nimmt die Container aller Datensatzarten.
	 *
	 * @param arcMgr           Der Archiv-Manager.
	 * @param objID            Die Objekt-ID.
	 * @param atgID            Die Attributgruppen-ID.
	 * @param aspID            Die Aspekt-ID.
	 * @param sv               Die Simulationsvariante.
	 * @param deletedTreatment Gibt an, ob geloeschte Container ignoriert werden [0], als 2 Datensaetze (Min/Max) zaehlen [1], oder voll gezaehlt werden [2]
	 * @param ownerTask        Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(ArchiveManager arcMgr, long objID, long atgID, long aspID, int sv, int deletedTreatment, Task ownerTask)
			throws PersistenceException {
		this(arcMgr, objID, atgID, aspID, sv, true, true, true, true, deletedTreatment, ownerTask);
	}

	/** @see #MultiContainerDataIterator(ArchiveManager,long,long,long,int,int,Task) mit deletedTreatment==0 */
	public MultiContainerDataIterator(ArchiveManager arcMgr, long objID, long atgID, long aspID, int sv, Task ownerTask) throws PersistenceException {
		this(arcMgr, objID, atgID, aspID, sv, true, true, true, true, DEL_IGNORE, ownerTask);
	}

	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 * <p/>
	 * Bei den Container-ID-Listen gilt: <ul> <li>Es werden genau die Container mit den angegebenen Container-IDs genommen</li> <li>Bei <code>null</code> werden
	 * alle Container genommen</li> <li>Bei {@link IndexResult#EMPTY} wird kein Container genommen</li> </ul>
	 *
	 * @param arcMgr    Der Archiv-Manager.
	 * @param objID     Die Objekt-ID.
	 * @param atgID     Die Attributgruppen-ID.
	 * @param aspID     Die Aspekt-ID.
	 * @param sv        Die Simulationsvariante.
	 * @param oaRes     Liste mit Container-IDs fuer Online-Aktuell erhaltene Datensaetze.
	 * @param onRes     Liste mit Container-IDs fuer Online-Nachgeliefert erhaltene Datensaetze.
	 * @param naRes     Liste mit Container-IDs fuer Nachgefordert-Aktuell erhaltene Datensaetze.
	 * @param nnRes     Liste mit Container-IDs fuer Nachgefordert-Nachgeliefert erhaltene Datensaetze.
	 * @param ownerTask Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr,
			long objID,
			long atgID,
			long aspID,
			int sv,
			IndexResult oaRes,
			IndexResult onRes,
			IndexResult naRes,
			IndexResult nnRes,
			Task ownerTask) throws PersistenceException {
		this(arcMgr, objID, atgID, aspID, sv, oaRes, onRes, naRes, nnRes, DEL_IGNORE, ownerTask);
	}

	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 * <p/>
	 * Bei den Container-ID-Listen gilt: <ul> <li>Es werden genau die Container mit den angegebenen Container-IDs genommen</li> <li>Bei <code>null</code> werden
	 * alle Container genommen</li> <li>Bei {@link IndexResult#EMPTY} wird kein Container genommen</li> </ul>
	 *
	 * @param arcMgr           Der Archiv-Manager.
	 * @param objID            Die Objekt-ID.
	 * @param atgID            Die Attributgruppen-ID.
	 * @param aspID            Die Aspekt-ID.
	 * @param sv               Die Simulationsvariante.
	 * @param oaRes            Liste mit Container-IDs fuer Online-Aktuell erhaltene Datensaetze.
	 * @param onRes            Liste mit Container-IDs fuer Online-Nachgeliefert erhaltene Datensaetze.
	 * @param naRes            Liste mit Container-IDs fuer Nachgefordert-Aktuell erhaltene Datensaetze.
	 * @param nnRes            Liste mit Container-IDs fuer Nachgefordert-Nachgeliefert erhaltene Datensaetze.
	 * @param deletedTreatment Gibt an, ob geloeschte Container ignoriert werden [0], als 2 Datensaetze (Min/Max) zaehlen [1], oder voll gezaehlt werden [2]
	 * @param ownerTask        Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr,
			long objID,
			long atgID,
			long aspID,
			int sv,
			IndexResult oaRes,
			IndexResult onRes,
			IndexResult naRes,
			IndexResult nnRes,
			int deletedTreatment,
			Task ownerTask) throws PersistenceException {

		this.deletedTreatment = deletedTreatment;
		this.ownerTask = ownerTask;
		this.arcMgr = arcMgr;
		this.persMgr = arcMgr.getPersMgr();
		didDir = new File(persMgr.getPathName(objID, atgID, aspID, sv));

		IndexResult[] idxRes = new IndexResult[]{oaRes, onRes, naRes, nnRes};
		for(int archiveDataKindIndex = OA; archiveDataKindIndex <= NN; archiveDataKindIndex++) {
			ArchiveDataKind archiveDataKind = Util.getDataKindFromIndex(archiveDataKindIndex);
			if(idxRes[archiveDataKindIndex] == null) {
				File adkPath = new File(persMgr.getPathName(objID, atgID, aspID, sv, archiveDataKindIndex));
				File[] contFiles = adkPath.listFiles(ContainerFile.CONT_FILENAME_FILTER);
				Util.sortFilesAlpha(contFiles); // Nach Namen (= Datenindexen, falls nicht NN) sortieren
				ContInfo[] contInfos = getContInfos(contFiles, archiveDataKind);
				ContainerWalk contWalk = new ContainerWalk(archiveDataKind);
				if(archiveDataKind.isRequested()) {
					contWalk.fillSorted(contInfos); // Nach Datenindex sortiert befuellen
				}
				else {
					contWalk.fillSimple(contInfos); // Nach Dateinamen = ContainerIndex = DatenIndex befuellen
				}
				contWalks[archiveDataKindIndex] = contWalk;
				getNextData(contWalk); // ContWalk haelt jetzt den ersten zu liefernden Datensatz bereit.
			}
			else if(!(idxRes[archiveDataKindIndex] == IndexResult.EMPTY)) {
				File adkPath = new File(persMgr.getPathName(objID, atgID, aspID, sv, archiveDataKindIndex));
				File[] contFiles = new File[idxRes[archiveDataKindIndex].size()];
				for(int i = 0; i < contFiles.length; i++) {
					contFiles[i] = new File(adkPath, PersistenceManager.getContFileName(idxRes[archiveDataKindIndex].getCID(i)));
				}
				Util.sortFilesAlpha(contFiles); // Nach Namen (= Datenindexen, falls nicht NN) sortieren
				ContInfo[] contInfos = getContInfos(contFiles, archiveDataKind);
				ContainerWalk contWalk = new ContainerWalk(archiveDataKind);
				if(archiveDataKind.isRequested()) {
					contWalk.fillSorted(contInfos);
				}
				else {
					contWalk.fillSimple(contInfos);
				}
				contWalks[archiveDataKindIndex] = contWalk;
				getNextData(contWalk); // ContWalk haelt jetzt den ersten zu liefernden Datensatz bereit.
			}
		}
	}

	/**
	 * Erzeugt einen Multi-Iterator, der ueber mehrere Container und Datensatzarten laeuft und dabei als naechsten DS stets den mit dem kleinsten Index liefert.
	 * <p/>
	 * Bei den Container-ID-Listen gilt: <ul> <li>Es werden genau die Container mit den angegebenen Container-IDs genommen</li> <li>Bei <code>null</code> werden
	 * alle Container genommen</li> <li>Bei {@link IndexResult#EMPTY} wird kein Container genommen</li> </ul>
	 *
	 * @param arcMgr           Der Archive-Manager.
	 * @param it               Die aktuelle Position dieses DataIdentTree-Iterators liefert die Datenidentifikation.
	 * @param oaRes            Liste mit Container-IDs fuer Online-Aktuell erhaltene Datensaetze.
	 * @param onRes            Liste mit Container-IDs fuer Online-Nachgeliefert erhaltene Datensaetze.
	 * @param naRes            Liste mit Container-IDs fuer Nachgefordert-Aktuell erhaltene Datensaetze.
	 * @param nnRes            Liste mit Container-IDs fuer Nachgefordert-Nachgeliefert erhaltene Datensaetze.
	 * @param deletedTreatment Gibt an, ob geloeschte Container ignoriert werden [0], als 2 Datensaetze (Min/Max) zaehlen [1], oder voll gezaehlt werden [2]
	 * @param ownerTask        Aufrufender Task. Wird bei Zugriffen auf die Persistenz benoetigt.
	 *
	 * @throws PersistenceException
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr,
			DataIdentTreeIterator it,
			IndexResult oaRes,
			IndexResult onRes,
			IndexResult naRes,
			IndexResult nnRes,
			int deletedTreatment,
			Task ownerTask) throws PersistenceException {
		this(arcMgr, it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV(), oaRes, onRes, naRes, nnRes, deletedTreatment, ownerTask);
	}

	/**
	 * @see #MultiContainerDataIterator(ArchiveManager,DataIdentTreeIterator,IndexResult,IndexResult,IndexResult,IndexResult,boolean,Task) mit deletedTreatment ==
	 *      0 (IGNORE)
	 */
	public MultiContainerDataIterator(
			ArchiveManager arcMgr, DataIdentTreeIterator it, IndexResult oaRes, IndexResult onRes, IndexResult naRes, IndexResult nnRes, Task ownerTask)
			throws PersistenceException {
		this(arcMgr, it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV(), oaRes, onRes, naRes, nnRes, DEL_IGNORE, ownerTask);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Oeffentliche Schnittstelle
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liefert Datensatz-Iterator zurueck, der auf dem naechsten Datensatz steht.
	 *
	 * @return Datensatz-Iterator mit naechstem Datensatz, <code>null</code> wenn alle Iteratoren am Ende sind
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator readNextDIdx() throws PersistenceException {

		// Zuerst den Container-Walk, der den letzten Datensatz geliefert hat, iterieren:
		if(lastContWalk != null) getNextData(lastContWalk);

		// Pruefen, ob nicht bereits alle Iteratoren am Ende sind:
		boolean anyLeft = false;
		for(int adk = OA; adk <= NN; adk++) {
			ContainerWalk contWalk = contWalks[adk];
			anyLeft |= (contWalk != null && !contWalk.done);
		}
		if(!anyLeft) return null; // Alle Iteratoren am Ende.

		// Container-Walk suchen, der naechsten Datensatz liefert:
		ContainerWalk nextContWalk = null;
		for(int adk = OA; adk <= NN; adk++) {
			ContainerWalk contWalk = contWalks[adk];
			if(contWalk != null && !contWalk.done) {
				if(nextContWalk == null) {
					nextContWalk = contWalk;
				}
				else if(contWalk.contInfo.iter.getDataIndex() < nextContWalk.contInfo.iter.getDataIndex()) nextContWalk = contWalk;
			}
		}

		// Gefundenen Container-Walk merken und zurueck liefern:
		lastContWalk = nextContWalk;
		return nextContWalk.contInfo.iter;
	}

	/**
	 * Liefert Datensatz-Iterator zurueck, der auf dem naechsten Datensatz steht, und ueberspringt diesen Datensatz.
	 *
	 * @return Datensatz-Iterator mit naechstem Datensatz
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator readSkipNextDIdx() throws PersistenceException {
		ContainerDataIterator cdi = readNextDIdx();
		if(cdi != null) cdi.skipData();
		return cdi;
	}

	/** Schliesst den Multi-Iterator und damit auch alle noch offenen Container-Dateien. */
	public void close() {
		for(ContainerWalk contWalk : contWalks) {
			if(contWalk != null) {
				contWalk.contInfo = null;
				contWalk.done = true;
				if(contWalk.currIdx > -1) leaveContainer(contWalk, contWalk.currIdx);
			}
		}
	}

	/**
	 * Liefert die Container-ID des Containers, der den letzten Datensatz geliefert hat.
	 *
	 * @return Container-ID des Containers, der den letzten Datensatz geliefert hat
	 */
	public long getCurContID() {
		return lastContWalk.contInfo.contId;
	}

	/**
	 * Liefert den ContainerFile-Index fuer {@link PersistenceManager#getContainerHeaderParamAsLong(Task,int,KeyValParam)} des Containers, der den letzten
	 * Datensatz geliefert hat.
	 *
	 * @return ContainerFile-Index des Containers, der den letzten Datensatz geliefert hat
	 */
	public int getCurContFileIdx() {
		return getContainerFileIdx(lastContWalk.dataKind, lastContWalk.contInfo.contInfoIdx);
	}

	/**
	 * Liefert eine textuelle Beschreibung der aktuellen Position aus Verzeichnis, Container und Datensatz.
	 *
	 * @return Textuelle Beschreibung der aktuellen Position
	 */
	public String descPosition() {
		String result = "DID-Verzeichnis: " + didDir.getAbsolutePath();
		if(lastContWalk != null) {
			result += ", Container: " + lastContWalk.contInfo.file + ", Datensatz mit Datenindex: " + lastContWalk.contInfo.iter.getDataIndex();
		}
		return result;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Methoden
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Sucht im angegebenen Container-Walk den Container mit den naechsten zu liefernden Datensatz und vermerkt diesen Container im Container-Walk.
	 *
	 * @param contWalk Zu iterierender Container-Walk
	 */
	private void getNextData(ContainerWalk contWalk) {
		if(contWalk.done) return; // Container-Walk ist bereits fertig

		if(contWalk.currIdx < 0) {
			// Wir starten gerade. Ersten ContWalk-Index (Bucket) adressieren:
			contWalk.currIdx = 0;
			if(contWalk.currIdx > contWalk.maxIdx()) {
				// ContWalk ist leer -> ContWalk ist erledigt
				contWalk.done = true;
				return;
			}
			// Im ersten ContWalk-Index (Bucket) alle Container zum Lesen oeffnen:
			if(!accessContainer(contWalk)) return; // Loggen und Schliessen ist bereits in accessContainer() passiert.
		}

		while(true) { // Schleife ueber alle Container-Walk-Indexe (Buckets)

			// Container, aus dem letzter Datensatz gelesen wurde, iterieren:
			if(contWalk.contInfo != null) {
				boolean success = false;
				try {
					success = contWalk.contInfo.iter.readNextDataHdr();
				}
				catch(PersistenceException e) {
					logger.error("Iterieren des ContainerDataIterator fehlgeschlagen.", e);
				}
				if(!success) leaveContainer(contWalk.contInfo);
			}

			ContInfo sendContainer = null; // Container, aus dem der naechste Datensatz zu liefern ist

			// Ueber alle unerledigten Container des aktuellen ContWalk-Idx (Bucket) gehen und kleinsten DS finden:
			for(ContInfo cont : contWalk.getContainer(contWalk.currIdx)) {
				if(!cont.done) {
					if(sendContainer == null) {
						sendContainer = cont;
					}
					else if(cont.iter.getDataIndex() < sendContainer.iter.getDataIndex()) sendContainer = cont;
				}
			}

			// Falls der naechste Datensatz gefunden worden ist:
			if(sendContainer != null) {
				contWalk.contInfo = sendContainer; // CW merkt sich Container mit naechstem Datensatz.
				return;
			}

			// In diesem Container-Walk-Index (Bucket) sind alle Container durchlaufen:
			leaveContainer(contWalk, contWalk.currIdx);
			if(contWalk.currIdx < contWalk.maxIdx()) {
				// Mit naechstem Container-Walk-Index (Bucket) fortfahren:
				contWalk.currIdx++;
				contWalk.contInfo = null; // Letzten gemerkten Container verwerfen.
				if(!accessContainer(contWalk)) return; // Loggen und Schliessen ist bereits in accessContainer() passiert.
			}
			else {
				// Alle Container-Walk-Indexe (Buckets) sind durchlaufen.
				contWalk.done = true;
				contWalk.contInfo = null;
				return;
			}
		}
	}

	/**
	 * Oeffnet im angegebenen Container-Walk alle Container innerhalb des aktuellen Container-Walk-Indexes zum Lesen. Der ContainerDataIterator wird in jedem
	 * Container bereits auf den ersten Datensatz positioniert.
	 *
	 * @param contWalk Container-Walk
	 *
	 * @return <code>true</code> falls Container zum Lesen geoeffnet werden koennen, <code>false</code> sonst
	 */
	private boolean accessContainer(ContainerWalk contWalk) {
		try {
			for(ContInfo contInfo : contWalk.getContainer(contWalk.currIdx)) {
				// Access:
				int cfIdx = getContainerFileIdx(contWalk.dataKind, contInfo.contInfoIdx); // ContainerFile-Index fuer Persistenz
				persMgr.accessContainerReadOnly(ownerTask, cfIdx, contInfo.file);
				contInfo.accessed = true;
				// Kontrolle ob geloescht:
				persMgr.readContainerHeader(ownerTask, cfIdx);
				contInfo.deleted = persMgr.getContainerHeaderParamAsBoolean(ownerTask, cfIdx, ContainerHdr.CHP_DELETED);
				// Iterator auf ersten Datensatz stellen
				if(contInfo.deleted) { // Container geloescht
					if(deletedTreatment == DEL_MINMAX) {
						contInfo.iter = persMgr.getHdrData(ownerTask, cfIdx);
						contInfo.iter.readNextDataHdr();
					}
					else if(deletedTreatment == DEL_FULLCOUNT) {
						contInfo.iter = persMgr.getArchiveData(ownerTask, cfIdx, true);
						contInfo.iter.readNextDataHdr();
					}
					else // IGNORE
					{
						leaveContainer(contInfo);
					}
				}
				else { // Kein geloeschter Container
					contInfo.iter = persMgr.getArchiveData(ownerTask, cfIdx);
					if(!contInfo.iter.readNextDataHdr()) leaveContainer(contInfo);
				}
			}
		}
		catch(PersistenceException e) {
			logger.error("Zugriff auf Container oder ContainerDataIterator fehlgeschlagen.", e);
			close(); // Abbruch
			return false;
		}
		return true;
	}

	/**
	 * Verlaesst im angegebenen Container-Walk alle Container innerhalb des angegebenen Container-Walk-Indexes.
	 *
	 * @param contWalk    Container-Walk
	 * @param contWalkIdx Container-Walk-Index
	 */
	private void leaveContainer(ContainerWalk contWalk, int contWalkIdx) {
		for(ContInfo contInfo : contWalk.getContainer(contWalkIdx)) {
			leaveContainer(contInfo);
		}
	}

	/**
	 * Verlaesst den durch die Container-Info spezifizierten Container und markiert ihn als erledigt.
	 *
	 * @param contInfo Zu verlassender Container
	 */
	private void leaveContainer(ContInfo contInfo) {
		// Datensatz-Iterator schliessen:
		try {
			if(contInfo.iter != null) contInfo.iter.close();
		}
		catch(PersistenceException e) {
			logger.error("Schliessen des ContainerDataIterator fehlgeschlagen.", e);
		}
		finally {
			contInfo.done = true;
		}
		// Container verlassen:
		try {
			if(contInfo.accessed) {
				int cfIdx = getContainerFileIdx(contInfo.dataKind, contInfo.contInfoIdx); // ContainerFile-Index fuer Persistenz
				persMgr.leaveContainer(ownerTask, cfIdx);
			}
		}
		catch(PersistenceException e) {
			logger.error("Verlassen des Containers fehlgeschlagen.", e);
		}
		finally {
			contInfo.accessed = false;
		}
	}

	/**
	 * Liefert die Container-Infos zu den uebergebenen Container-Dateien der angegebenen Datensatzart.
	 *
	 * @param contFiles Liste mit Container-Dateien
	 * @param dataKind  Datensatzart der Container
	 *
	 * @return Liste mit Container-Infos
	 *
	 * @throws PersistenceException
	 */
	private ContInfo[] getContInfos(File[] contFiles, ArchiveDataKind dataKind) throws PersistenceException {
		if(contFiles == null) return new ContInfo[0];

		ContInfo[] contInfos = new ContInfo[contFiles.length];
		for(int i = 0; i < contFiles.length; i++) {
			try {
				File contFile = contFiles[i];                         // Container-Datei
				long contId = PersistenceManager.getContID(contFile); // Container-ID
				persMgr.accessContainerReadOnly(ownerTask, contFile);
				persMgr.readContainerHeader(ownerTask);
				long minDataIdx = persMgr.getContainerHeaderParamAsLong(ownerTask, ContainerHdr.CHP_DATA_IDX_MIN);
				long maxDataIdx = persMgr.getContainerHeaderParamAsLong(ownerTask, ContainerHdr.CHP_DATA_IDX_MAX);
				if(minDataIdx == 0 && maxDataIdx == 0) {
					// Nicht abgeschlossenen Container behandeln:
					long[] dataIdx = getDataIdx(contFile, contId, dataKind);
					minDataIdx = dataIdx[0];
					maxDataIdx = dataIdx[1];
				}
				contInfos[i] = new ContInfo(contId, contFile, dataKind, minDataIdx, maxDataIdx);
			}
			finally {
				persMgr.leaveContainer(ownerTask);
			}
		}
		return contInfos;
	}

	/** Eindeutiger ContainerFile-Index fuer Persistenz-Zugriff. */
	private int getContainerFileIdx(ArchiveDataKind dataKind, int contInfoIdx) {
		int dataKindOffset = dataKind.getCode();
		int containerOffset = contInfoIdx * Util.getAllDataKinds().length;
		return dataKindOffset + containerOffset;
	}

	/**
	 * Bei nicht abgeschlossenen Containern (min/max DataIdx steht noch nicht im Header) liefert diese Methode min/max DataIdx entweder aus den Container-Indexen
	 * oder aus den Datensaetzen des Containers, falls die Container-Indexe diese Information nicht vorhalten. persMgr.accessContainerReadOnly(ownerTask, contFile)
	 * muss bereits vom Aufrufer dieser Methode ausgefuehrt worden sein. persMgr.leaveContainer(ownerTask) muss vom Aufrufer dieser Methode aufgerufen werden,
	 * nachdem diese Methode returniert ist.
	 *
	 * @param contFile Container-Datei
	 * @param contId   Container-ID
	 * @param dataKind Datensatzart des Containers
	 *
	 * @return Array mit min und max DataIdx
	 *
	 * @throws PersistenceException
	 */
	private long[] getDataIdx(File contFile, long contId, ArchiveDataKind dataKind) throws PersistenceException {
		long[] dataIdx = new long[2]; // min/max DataIdx
		long di[] = PersistenceManager.extractDIdFromPath(contFile.getParentFile()); // DID

		// 1. Versuch: Container-Indexe anfragen:
		DataIdentNode din = arcMgr.getDataIdentTree().existsCreate(di[0], di[1], di[2], (int)di[3]); // DIN not null
		boolean criticalAreaEntered = false;
		try {
			criticalAreaEntered = din.enterCriticalArea();
			if(!criticalAreaEntered) {
				logger.warning("Synchronisierter Index-Zugriff fehlgeschlagen fuer DIN " + din);
				throw new PersistenceException("Synchronisierter Index-Zugriff fehlgeschlagen fuer DIN " + din);
			}
			if(din.indexesAssigned(dataKind)) {
				if(din.getOpenContID(dataKind) == contId) {
					// Container-Indexe haben Info:
					dataIdx[0] = din.minDIdx(dataKind);
					dataIdx[1] = din.maxDIdx(dataKind);
					return dataIdx;
				}
			}
		}
		catch(IndexException ie) {
			logger.warning("Index-Zugriff fehlgeschlagen fuer DIN " + din, ie);
			throw new PersistenceException("Index-Zugriff fehlgeschlagen fuer DIN " + din, ie);
		}
		finally {
			if(din != null && criticalAreaEntered) din.exitCriticalArea();
		}

		// 2. Versuch: Ueber Datensaetze des Containers laufen:
		long minDataIdx = Long.MAX_VALUE;
		long maxDataIdx = 0;
		ContainerDataIterator iter = persMgr.getArchiveData(ownerTask);
		while(iter.readNextDataHdr()) {
			long dIdx = iter.getDataIndex();
			if(dIdx < minDataIdx) minDataIdx = dIdx;
			if(dIdx > maxDataIdx) maxDataIdx = dIdx;
			iter.skipData();
		}
		dataIdx[0] = minDataIdx;
		dataIdx[1] = maxDataIdx;
		return dataIdx;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Klasse ContainerWalk
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	private class ContainerWalk {

		private List<List<ContInfo>> cw = new ArrayList<List<ContInfo>>();

		private ArchiveDataKind dataKind;

		private boolean done = false;

		private int currIdx = -1;

		/** Iterator ueber die Datensaetze des aktuell auszulesenden Containers. */
		private ContInfo contInfo = null;

		private ContainerWalk(ArchiveDataKind dataKind) {
			this.dataKind = dataKind;
		}

		/**
		 * Liefert Liste aller Container-Infos, die beim angegebenen Container-Walk-Index abgelegt sind. Sollte eine Liste am Ende des ContainerWalks abgefragt
		 * werden, die noch nicht existiert, wird sie erzeugt und angehangen.
		 *
		 * @param cwIdx Container-Walk-Index
		 *
		 * @return Liste von Container-Infos
		 */
		private List<ContInfo> getContainer(int cwIdx) {
			List<ContInfo> container = null;
			try {
				container = cw.get(cwIdx);
			}
			catch(IndexOutOfBoundsException e) {
			}
			if(container == null) {
				assert (cw.size() == cwIdx) : "Versuch ContWalk-Idx " + cwIdx + " zu fuellen bei ContWalk-Laenge " + cw.size();
				container = new ArrayList<ContInfo>();
				cw.add(cwIdx, container);
			}
			return container;
		}

		/**
		 * Fuellt die angegebenen ContInfos in ihrer Reihenfolge in den ContWalk, jeweils einen pro ContWalk-Idx.
		 *
		 * @param contInfos In den ContWalk zu fuellende ContInfos.
		 */
		private void fillSimple(ContInfo[] contInfos) {
			for(int idx = 0; idx < contInfos.length; idx++) {
				getContainer(idx).add(0, contInfos[idx]); // Einen Container pro ContWalk-Idx einfuellen
			}
		}

		/**
		 * Fuellt die angegebenen ContInfos nach DatenIndex sortiert in den ContWalk, jeweils sich ueberschneidende in einen ContWalk-Idx.
		 *
		 * @param contInfos In den ContWalk zu fuellende ContInfos.
		 */
		private void fillSorted(ContInfo[] allCont) {
			int cwIdx = 0; // Start
			while(!allInContWalk(allCont)) {
				// Noch nicht in ContWalk einsortierte ContInfo mit kleinstem min Wert suchen:
				ContInfo contMin = null;
				for(ContInfo cont : allCont) {
					if(cont.contWalkIdx < 0 && (contMin == null || cont.minDataIdx < contMin.minDataIdx)) contMin = cont;
				}

				// Die gefundene ContInfo mit kleinstem min Wert in ContWalk einsortieren;
				// sie ist die erste ContInfo im gerade zu fuellenden ContWalk-Index:
				contMin.contWalkIdx = cwIdx;
				contMin.contInfoIdx = 0;
				getContainer(cwIdx).add(0, contMin);

				// Alle nicht zugeordneten ContInfos suchen, die sich mit dem gefundenen ueberschneiden:
				long min = contMin.minDataIdx, max = contMin.maxDataIdx;
				boolean proceed = true;
				while(proceed) {
					proceed = false;
					for(ContInfo cont : allCont) {
						if(cont.contWalkIdx < 0) {
							if(cont.minDataIdx <= max && cont.maxDataIdx >= min) { // Kriterium fuer Ueberschneidung
								// Gefundene ContInfo in ContWalk einfuegen:
								cont.contWalkIdx = cwIdx;
								cont.contInfoIdx = getContainer(cwIdx).size();
								getContainer(cwIdx).add(cont.contInfoIdx, cont);
								// Nochmal suchen, da ein anderer Container sich mit dem gerade gefundenen ueberschneiden koennte
								min = Math.min(min, cont.minDataIdx);
								max = Math.max(max, cont.maxDataIdx);
								proceed = true;
							}
						}
					}
				}
				// Dieser ContWalkIdx ist gefuellt. Mit naechstem weiter machen:
				cwIdx++;
			}
		}

		private boolean allInContWalk(ContInfo[] contInfos) {
			for(ContInfo contInfo : contInfos) {
				if(contInfo.contWalkIdx < 0) return false; // ContInfo gefunden, die nicht im ContWalk ist
			}
			return true;
		}

		private int maxIdx() {
			return cw.size() - 1;
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Klasse ContInfo
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Kapselt folgende Daten zu einem Container: Datei, Container-ID, Datensatzart und min/max Datenindex. Haelt darueber hinaus Informationen zum Iterieren ueber
	 * den Container.
	 */
	private class ContInfo {

		private long contId, minDataIdx, maxDataIdx;

		private File file;

		private ArchiveDataKind dataKind;

		private boolean accessed = false, done = false, deleted = false;

		/** Container-Walk-Index, an dem diese Container-Info in den Container-Walk einsortiert worden ist. */
		private int contWalkIdx = -1;

		/** Index, an dem diese Container-Info innerhalb ihres Container-Walk-Indexes einsortiert worden ist. */
		private int contInfoIdx = -1;

		/** Iterator ueber die Datensaetze des aktuell auszulesenden Containers. */
		private ContainerDataIterator iter;

		private ContInfo(long contId, File file, ArchiveDataKind dataKind, long minDataIdx, long maxDataIdx) {
			this.contId = contId;
			this.file = file;
			this.dataKind = dataKind;
			this.minDataIdx = minDataIdx;
			this.maxDataIdx = maxDataIdx;
		}
	}
}
